/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

#ifdef rtaoSkySample
    /* - */
#endif

vec3 getRTAO(vec3 viewPos, vec3 viewNormal, float dither, bool hand, float skyOcclusion) {
    //if (hand) return 1.0;

    const uint samples  = rtaoSPP;
    const uint steps    = rtaoSteps;

    vec3 viewDir        = normalize(viewPos);

    #if (defined rtaoSkySample || !defined DIM)
        vec3 ambient        = vec3(0.0);
    #else
        float ambient       = 0.0;
    #endif

    const float maxRayDist  = rtaoLength;
    const float maxStepLength  = maxRayDist / float(steps);
    float distCoeff     = mix(0.125, 1.0, linStep(length(viewPos), maxStepLength * 0.125, maxRayDist));
    float startStepLength = maxStepLength * distCoeff;
    float stepLength    = startStepLength;

    vec3 rStart         = viewPos;
    vec3 rPos           = rStart;

    bool hit    = false;
    bool hitBehind = false;
    float hitErrorScale = 1.0;

    vec3 hitPos = vec3(0.0);

    const float refineMult  = pow(0.5, float(steps));

    vec3 upDir  = mat3(gbufferModelView) * vec3(0.0, 1.0, 0.0);

    float noise     = ditherBluenoise();
    //float noise     = fract(hash12(gl_FragCoord.xy / viewSize) + float(frameCounter) / pi);

    for (uint i = 0; i < samples; ++i) {
        vec3 rDir           = sphereMap(fermatsSpiralGoldenN(float(i + 1) * 64.0 + noise * 64.0, float(samples) * 64.0));
            rDir            = mat3(gbufferModelView) * rDir;

        #ifndef DIM
            rDir            = mix(rDir, upDir, 0.33);
        #endif

        if (dot(rDir, viewNormal) < 0.0) rDir = -rDir;

        vec3 rStep  = rDir * startStepLength;

        float maxPos2D  = max2(abs(rStart.xy + rStep.xy));

        if (maxPos2D >= 1.0) {
            float coeff     = clamp((1.0 / (maxPos2D + 1e-2)) * 0.5 + 0.5, 0.1, 0.99);
            stepLength      = startStepLength * coeff;
            rStep           = rDir * stepLength;
        }

        rPos    = rStart + rStep * dither;

        vec3 screenPos  = viewToScreenSpace(rPos);

        vec3 worldDir       = mat3(gbufferModelViewInverse) * rDir;
        vec2 sphereCoord    = unprojectSphere(worldDir);
            sphereCoord    *= rcp(skyCaptureResolution);
            sphereCoord.y  += exp2(-SKY_RENDER_LOD) * 2.0;

        #if (defined rtaoSkySample || !defined DIM)
            vec3 fakeBounce     = ((1.0 - expf(-max0(-worldDir.y - 0.005) * pi)) * 0.04 + 0.01) * directLightColor;
        #endif

        if (saturate(screenPos.xy) != screenPos.xy){
            #ifdef rtaoSkySample
                ambient += (texture(colortex5, sphereCoord).rgb * sqrt3 + fakeBounce) * saturate(dot(viewNormal, rDir));
            #else
                #ifndef DIM
                    ambient += (indirectLightCol + fakeBounce) * saturate(dot(viewNormal, rDir));
                #else
                    ambient += saturate(dot(viewNormal, rDir));
                #endif
            #endif

            continue;
        }

        float depth     = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        vec3 sampleViewPos = screenToViewSpace(vec3(screenPos.xy, depth));

        float error     = distance(rPos, sampleViewPos);

        float slength   = length(rStep);

        float delta     = rPos.z - sampleViewPos.z;
        if (dot(rDir, viewDir) > 0.0) delta = -delta;

        bool potentialHit = delta <= 0.0 || distance(rStart, rPos) > 0.1;

        if (error < stepLength && potentialHit) {
            hitBehind = screenPos.z >= depth;
            hitErrorScale = (hitBehind ? stepLength - error : error) * rcp(stepLength + 1e-10);
            hitPos = rPos;
            //break;

            vec3 refineStep = rStep * hitErrorScale * 0.85;
            vec3 refinePos  = hitPos;
            if (hitBehind) refinePos -= rStep;

            vec3 refineHitPos = hitPos;

            for (uint j = 0; j < steps; ++j) {
                vec3 screenPos  = viewToScreenSpace(refinePos);

                if (saturate(screenPos.xy) != screenPos.xy) break;

                float depth     = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
                vec3 sampleViewPos = screenToViewSpace(vec3(screenPos.xy, depth));

                float error     = distance(rPos, sampleViewPos);

                if (screenPos.z > depth) refinePos -= refineStep;
                else refineHitPos = refinePos;

                //float rayDotUp  = dot(normalize(sampleViewPos - rStart), upvecView);

                if (screenPos.z > depth && distance(screenPos.z, depth) < stepLength) {
                    hit = true;
                    break;
                }
                refineStep *= 0.5;
                refinePos += refineStep;
            }
        }

        if (hit) continue;

        #ifdef rtaoSkySample
            ambient += (texture(colortex5, sphereCoord).rgb * sqrt3 + fakeBounce) * saturate(dot(viewNormal, rDir));
        #else
            #ifndef DIM
                ambient += (indirectLightCol + fakeBounce) * saturate(dot(viewNormal, rDir));
            #else
                ambient += saturate(dot(viewNormal, rDir));
            #endif
        #endif
    }
    ambient    /= float(samples);
    ambient     = max0(ambient);

    #ifndef DIM
        ambient *= cube(skyOcclusion);
    #endif

    #if (defined rtaoSkySample || !defined DIM)
        return ambient;
    #else
        return indirectLightCol * ambient;
    #endif
}