/*
====================================================================================================

    Copyright (C) 2020 RRe36

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://rre36.github.io/license/

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/

struct ssrtData {
    vec3 bounce;
    vec4 emission;
};

ssrtData SSRT(vec3 viewPos, vec3 sceneNormal, float dither, bool hand) {
    const uint samples  = ssrtSPP;

    vec3 viewDir    = normalize(viewPos);
    vec3 viewNormal = mat3(gbufferModelView) * sceneNormal;
    vec3 scenePos   = viewToSceneSpace(viewPos);

    vec3 bounced    = vec3(0.0);
    vec4 emission   = vec4(0.0);

    const float maxRayDist  = 2.61 * ssrtRange;
    const float rayLength   = maxRayDist;
    const float falloffScale = 2.61 / maxRayDist;

    vec3 rayStart   = viewPos;
    vec3 rayPos     = rayStart;

    bool hit        = false;

    vec3 hitPos     = vec3(0.0);

    float noise     = ditherBluenoise();
    //float noise     = fract(hash12(gl_FragCoord.xy / viewSize) + float(frameCounter) / pi);

    #ifdef rsmBounce
        vec3 shadowPos      = scenePos + vec3(0.08) * lightvec;
            shadowPos       = viewMAD(shadowModelView, shadowPos);
        vec3 shadowClipPos  = projMAD(shadowProjection, shadowPos);

        vec2 rsmDir     = sincos(ditherSize * goldenAngle);

        vec2 ditherNoise = vec2(dither, noise);
    #endif

    for (uint i = 0; i < samples; ++i) {
        vec3 rayDir = sphereMap(fermatsSpiralGoldenN(float(i + 1) * 64.0 + noise * 64.0, float(samples) * 64.0));
        vec3 rayDirScene = rayDir;
            rayDir  = mat3(gbufferModelView) * rayDir;

        if (dot(rayDir, viewNormal) < 0.0) {
            rayDir = -rayDir;
            rayDirScene = -rayDirScene;
        }

        rayPos      = rayStart + rayDir * rayLength * dither;

        vec3 screenPos  = viewToScreenSpace(rayPos);

        if (saturate(screenPos.xy) != screenPos.xy){
            #ifdef rsmBounce
                bounced    += rsmBounceResponse(shadowClipPos, shadowPos, sceneNormal, ditherNoise, i);
            #endif
            
            continue;
        }

        float depth     = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        vec3 sampleViewPos = screenToViewSpace(vec3(screenPos.xy, depth));

        float error     = distance(rayPos, sampleViewPos);

        float stepLength = rayLength;

        float delta     = rayPos.z - sampleViewPos.z;
        if (dot(rayDir, viewDir) > 0.0) delta = -delta;

        bool potentialHit = delta <= 0.0 || distance(rayStart, rayPos) > 0.1;

        ivec2 pixelPos  = ivec2(screenPos.xy * viewSize);

        float distFromStart = distance(sampleViewPos, rayStart);

        if (error < stepLength && potentialHit && depth < 1.0 && distFromStart < maxRayDist) {
            vec3 tex2   = texelFetch(colortex2, pixelPos, 0).rgb;

            vec3 normal = texelFetch(colortex6, pixelPos, 0).xyz * 2.0 - 1.0;

            vec3 direction  = normalize(sampleViewPos - rayStart);

            vec3 albedo     = texelFetch(colortex0, pixelPos, 0).rgb;

            float lambert   = diffuseLambert(viewNormal, direction);

            if (dot(sceneNormal, normal) > 0.75){
                #ifdef rsmBounce
                    bounced    += rsmBounceResponse(shadowClipPos, shadowPos, sceneNormal, ditherNoise, i);
                #endif
            } else {
                vec4 currBounce     = vec4(decode2x8(tex2.x), decode2x8(tex2.y));
                if (currBounce.a < 1e-2) continue;

                bounced    += (albedo * currBounce.yzw) * (currBounce.x * lambert * sqrt3);
            }

            #ifdef ssrtEmissionToggle
                if (tex2.z < 1e-2) continue;

                float albedoLum     = getLuma(albedo);

                vec3 emissionColor = saturate(mix((albedoLum) * sqr(normalize(albedo)), albedo, albedoLum));

                float emissionFalloff = expf(-max0(distFromStart - 0.33 * tex2.z) * falloffScale * 0.25);

                emission.rgb   += emissionColor * lambert * tex2.z * emissionFalloff;

                if ((tex2.z * lambert) > 0.0) emission.a += 1.0 * emissionFalloff;
            #endif

        } else {

            #ifdef rsmBounce
                bounced    += rsmBounceResponse(shadowClipPos, shadowPos, sceneNormal, ditherNoise, i);
            #endif

        }
    }

    bounced    /= float(samples);
    emission.rgb /= float(samples);
    emission.a  = saturate(emission.a);

    return ssrtData(bounced * pi, emission);
}

vec4 SSRT_Emission(vec3 viewPos, vec3 sceneNormal, float dither, bool hand) {
    const uint samples  = ssrtSPP;

    vec3 viewDir    = normalize(viewPos);
    vec3 viewNormal = mat3(gbufferModelView) * sceneNormal;

    vec3 bounced    = vec3(0.0);
    vec4 emission   = vec4(0.0);

    const float maxRayDist  = 2.61 * ssrtRange;
    const float rayLength   = maxRayDist;
    const float falloffScale = 2.61 / maxRayDist;

    vec3 rayStart   = viewPos;
    vec3 rayPos     = rayStart;

    bool hit        = false;

    vec3 hitPos     = vec3(0.0);

    float noise     = bayer32();

    for (uint i = 0; i < samples; ++i) {
        vec3 rayDir = sphereMap(fermatsSpiralGoldenN(float(i + 1) * 64.0 + noise * 64.0, float(samples) * 64.0));
        vec3 rayDirScene = rayDir;
            rayDir  = mat3(gbufferModelView) * rayDir;

        if (dot(rayDir, viewNormal) < 0.0) rayDir = -rayDir;

        rayPos      = rayStart + rayDir * rayLength * dither;

        vec3 screenPos  = viewToScreenSpace(rayPos);

        if (saturate(screenPos.xy) != screenPos.xy){
            continue;
        }

        float depth     = texelFetch(depthtex0, ivec2(screenPos.xy * viewSize), 0).x;
        vec3 sampleViewPos = screenToViewSpace(vec3(screenPos.xy, depth));

        float error     = distance(rayPos, sampleViewPos);

        float stepLength = rayLength;

        float delta     = rayPos.z - sampleViewPos.z;
        if (dot(rayDir, viewDir) > 0.0) delta = -delta;

        bool potentialHit = delta <= 0.0 || distance(rayStart, rayPos) > 0.1;

        ivec2 pixelPos  = ivec2(screenPos.xy * viewSize);

        float distFromStart = distance(sampleViewPos, rayStart);

        if (error < stepLength && potentialHit && depth < 1.0 && distFromStart < maxRayDist) {
            vec3 tex2   = texelFetch(colortex2, pixelPos, 0).rgb;

            vec3 normal = texelFetch(colortex6, pixelPos, 0).xyz * 2.0 - 1.0;

            vec3 direction  = normalize(sampleViewPos - rayStart);

            vec3 albedo     = texelFetch(colortex0, pixelPos, 0).rgb;

            float lambert   = diffuseLambert(viewNormal, direction);

            if (tex2.z < 1e-2) continue;

            float albedoLum     = getLuma(albedo);

            vec3 emissionColor = saturate(mix((albedoLum) * sqr(normalize(albedo)), albedo, albedoLum));

            float emissionFalloff = expf(-max0(distFromStart - 0.33) * falloffScale * 0.25);

            emission.rgb   += emissionColor * lambert * tex2.z * emissionFalloff;

            if ((tex2.z * lambert) > 0.0) emission.a += 1.0 * emissionFalloff;
        }
    }

    emission.rgb /= float(samples);
    emission.a  = saturate(emission.a);

    return emission;
}